import copy
from typing import Union
from funcy import collecting
from populate import populate_string, indent_to
import json
import shutil
import requests
import fire
from skema.infer import from_jsonschema
from .tree import Node, is_leaf_key
from skema.to_jsonschema import to_jsonschema
from typing import *
from populate import populate_string, indent_to, render_dict
from .constants import *
from .split_tree_parts import split_tree_parts
from .split_references import is_big_list, replace_occurrences, stronger_type, get_leaves
from functools import reduce
from .support import (capitalize, is_and_key, is_enum_key, is_key, is_list_key, topological_sort,
                      is_object, is_or_key, is_scalar, is_leaf, is_leaf_key, is_and_object)



imports = """
from typing import Any, Optional, List, Union, Callable
from typing_extensions import Literal
import skema
import fastjsonschema
from jsonschema import validate
from prtty import prettify
lmap = lambda func: lambda xs: list(map(func, xs))

class dotdict(dict):
    def __getattr__(self, name):
        try:
            return dict.__getitem__(self, name)
        except KeyError:
            return object.__getattribute__(self, name)
    @classmethod
    def from_dict(cls, obj: dict):
        assert isinstance(obj, dict)
        return cls(**obj)
    def validate(self):
        # return self.validate_(self)
        return validate(instance=self, schema=self._schema)
    @classmethod
    def fake(cls, resolvers={}):
        return cls(**(skema.fake_data(cls._schema, amount=1, from_json=True, resolvers=resolvers)[0] or {}))
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    def __repr__(self):
        return f'{self.__class__.__name__}({prettify(self)})'
"""

def is_circular(schema):
    try:
        json.dumps(schema)
    except Exception:
        return True

def get_local_schema(schema, typename,):
    try:
        schema = to_jsonschema(schema, ref=typename, resolve=True)
        if is_circular(schema):
            return to_jsonschema(schema, ref=typename, resolve=False)
        else:
            return schema
    except Exception as e:
        print(e)
        return {}

map_simple_types = dict( # TODO i am mapping from graphql types
    String='str',
    Str='str',
    str='str',
    Int='int',
    Float='float',
    null='None',
    Bool='bool',
    Boolean='bool',
    Any='Any',
    Json='Any',
)
def map_type(node: Node):
    if node.value in [OR]:
        raise Exception('non so che fare')
    key = node.children[0].value
    if key in map_simple_types:
        return map_simple_types[key]
    else:
        if key == OR:
            return handle_union(node.children[0])
        elif key == LIST:
            return f'List[{map_type(node.children[0])}]'
        return "'" + key + "'"

def get_initializer(node: Node):
    key = node.children[0].value
    if key in map_simple_types:
        return None
    else:
        if key == OR:
            if any(['"' in c.value for c in node.children[0].children]):
                return None
            return f'dotdict'
        elif key == LIST:
            return f'lmap({get_initializer(node.children[0]) or "lambda x: x"})'
        elif key == ANY:
            return None
        return f'{key}.from_dict'


def is_valid_as_reference(key: Node): # for graphql
    def is_valid_key(key):
        if is_or_key(key):
            return True
        if not is_key(key):
            return False
        list_node = key.children[0]
        return not is_big_list(list_node)
    # true se è un oggetto con solo leaf_key oppure con piccole liste come figli
    if is_object(key) and not is_list_key(key) and all([is_leaf_key(c) or is_valid_key(c) for c in key.children]):
        return True
    if is_and_key(key) or is_and_object(key):
        return True
    if is_big_list(key):
        return True
    return False


def handle_union(node: Node):
    options = ["'" + c.value + "'" if not '"' in c.value else c.value for c in node.children]
    options = [f'Literal[{n}]' if '"' in n else n for n in options]
    return f'Union[{", ".join(options)}]'



# def get_dependencies(node: Node):
#     global scalars
#     leaves = get_leaves(node)
#     scalars = [s.lower() for s in scalars]
#     def is_reference(x: Node):
#         return x.value.lower() not in scalars and not '"' in x.value
#     return set(x for x in leaves if is_reference(x) and x.value != node.value)


# def move_dependencies_up(refs):
#     refs = [(r, get_dependencies(r)) for r in refs]
#     print(refs)
#     return topological_sort(refs)
#     # return [x for x, _ in refs]


def merge_scalars(references):
    scalars = [STR, STRING, INT, FLOAT, REGEX, ANY, NULL, BOOL, 'Json']
    to_delete = {}
    for node in references:
        is_scalar_leafe = (is_key(node) and node.children[0].value in scalars)
        if is_scalar_leafe:
            new_type = map_type(node)
            obj = {node.value: new_type}
            # print('new_type', obj)
            to_delete.update(obj)
    # print('to_delete', to_delete)
    for ref in references:
        replace_occurrences(ref, to_delete)
    return [r for r in references if not r.value in to_delete.keys()]


def to_python(schema, hide=[], only=None):
    # get the main objects names from schema
    # generate a resolved jsonschema for each one
    refs = split_tree_parts(schema, 'python', is_valid_as_reference=is_valid_as_reference,  hide=hide, only=only)
    refs = list(reversed(refs))
    refs = merge_scalars(refs)
    # refs = move_dependencies_up(refs)
    # refs = [x for x in refs if not is_leaf_key(x)]
    print('\n\n'.join([str(x) for x in refs]))
    string = imports
    string += populate_string(exports, dict(
        all=[n.value for n in refs],
    ))
    for node in refs:
        typename = node.value
        if is_leaf_key(node):
            string += f'{typename} = {map_type(node)}' + '\n'
            continue
        if node.children[0].value == OR:
            string += f'{typename} = ' + handle_union(node.children[0]) + '\n'
            string += f'{typename}.from_dict = lambda x: x' + '\n'
            continue
        args = {c.value: not c.required for c in node.children}
        hints = {c.value: map_type(c) for c in node.children}
        setters = {c.value: get_initializer(c) for c in node.children}
        # print(local_schema)
        string += populate_string(
            template,
            dict(
                hints=hints,
                args=args,
                setters=setters,
                typename=typename,
                render_args=render_args,
                render_setters=render_setters,
                render_hints=render_hints,
                indent_to=indent_to,
                render_dict=render_dict,
            ),
        )
        string += '\n'
    for node in refs:
        typename = node.value
        local_schema = get_local_schema(schema, typename)
        string += populate_string(
            bottom,
            dict(
                schema=local_schema,
                typename=typename,
                render_dict=render_dict,
            ),
        )
    return string

exports = '''
__all__ = (
    ${{ indent_to('    ', '\\n'.join(['"' + x + '",' for x in all])) }}
)
'''

bottom = '''
${{typename}}._schema = ${{ render_dict(schema) }}
# ${{typename}}.validate_ = staticmethod(fastjsonschema.compile(${{typename}}._schema))

'''
template = """
class ${{typename}}(dotdict):
    _schema: dict
    # validate_: staticmethod

    ${{indent_to('    ', render_hints(hints, args)) + '\\n'}}

    def __init__(
        self, 
        *, 
        ${{indent_to('        ', render_args(args, hints)) }}, 
        **kwargs
    ):
        super().__init__(
            ${{indent_to('            ', render_setters(setters, args))}},
            **kwargs
        )



"""


def render_hints(hints: Dict[str, str], args: Dict[str, str]):
    hints = [
        f"{name}: Optional[{type}]" if args[name] else f"{name}: {type}"
        for name, type in hints.items()
    ]
    return "\n".join(hints)


def render_args(fields: Dict[str, bool], hints: Dict[str, str], default=lambda k: "None"):
    required_fields = [x for x, is_optional in fields.items() if not is_optional]
    optional_fields = [x for x, is_optional in fields.items() if is_optional]
    required_args = [f"{name}: {hints[name]}" for name in required_fields]
    optional_args = [f"{name}: {hints[name]}={default(name)}" for name in optional_fields]
    return ",\n".join(required_args + optional_args)


def render_setters(setters: Dict[str, str], args: Dict[str, bool]):
    args = [f"{name}={(initializer + '(' + name + ')' + f' if {name} != None else {name}') if initializer else name}" for name, initializer in setters.items()]
    return ",\n".join(args)


if __name__ == "__main__":
    s = imports
    s += populate_string(
        template,
        dict(
            render_args=render_args,
            render_hints=render_hints,
            typename="Classe",
            hints=dict(a="str", b="int"),
            args=dict(a=True, b=False),
            indent_to=indent_to,
            schema={},
            render_dict=render_dict,
        ),
    )
    # print(s)
    with open('example__.py', 'w') as f:
        f.write(s)




#
# Copy and pasta from PR https://github.com/HypothesisWorks/hypothesis-python/pull/393/files
#

# coding=utf-8
#
# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis-python
#
# Most of this work is copyright (C) 2013-2016 David R. MacIver
# (david@drmaciver.com), but it contains contributions by others. See
# CONTRIBUTING.rst for a full list of people who may hold copyright, and
# consult the git log if you need to determine who owns an individual
# contribution.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.
#
# END HEADER

from __future__ import division, print_function, absolute_import

import re
import sys
import string
from itertools import chain

import hypothesis.strategies as strats
from hypothesis._strategies import defines_strategy
from hypothesis.searchstrategy import SearchStrategy
from hypothesis.internal.compat import hunichr, hrange


def strategy_concat(strategies):
    """given a list of strategies yielding strings return a strategy yielding
    their concatenation."""
    return strats.tuples(*strategies).map(lambda x: u"".join(x))


class RegexStrategy(SearchStrategy):
    """Strategy for generating strings matching a regular expression.

    Currently does not support \B, positive and negative lookbehind
    assertions and conditional matching.

    """


    CATEGORY_DIGIT = string.digits
    CATEGORY_NOT_DIGIT = string.ascii_letters + string.punctuation
    CATEGORY_SPACE = string.whitespace
    CATEGORY_NOT_SPACE = string.printable.strip()
    CATEGORY_WORD = string.ascii_letters + string.digits + '_'
    CATEGORY_NOT_WORD = ''.join(set(string.printable)
                                .difference(string.ascii_letters +
                                            string.digits + '_'))

    def __init__(self, pattern):
        parsed = re.sre_parse.parse(pattern)
        self.cache = {}
        self.strategies = [self._handle_state(state) for state in parsed]

    def do_draw(self, data):
        self.cache = {}
        temp = [data.draw(strat) for strat in self.strategies]
        return "".join(temp)

    def _categories(self, category):
        if category == re.sre_parse.CATEGORY_DIGIT:
            return self.CATEGORY_DIGIT
        elif category == re.sre_parse.CATEGORY_NOT_DIGIT:
            return self.CATEGORY_NOT_DIGIT
        elif category == re.sre_parse.CATEGORY_SPACE:
            return self.CATEGORY_SPACE
        elif category == re.sre_parse.CATEGORY_NOT_SPACE:
            return self.CATEGORY_NOT_SPACE
        elif category == re.sre_parse.CATEGORY_WORD:
            return self.CATEGORY_WORD
        elif category == re.sre_parse.CATEGORY_NOT_WORD:
            return self.CATEGORY_NOT_WORD
        else:
            raise NotImplementedError("Unable to find category: {0}".format(category))

    def _handle_character_sets(self, state):
        opcode, value = state
        if opcode == re.sre_parse.RANGE:
            return [hunichr(val) for val in hrange(value[0], value[1] + 1)]
        elif opcode == re.sre_parse.LITERAL:
            return [hunichr(value)]
        elif opcode == re.sre_parse.CATEGORY:
            return self._categories(value)
        else:
            raise NotImplementedError("Unable to find character set: {0}".format(opcode))

    def _handle_state(self, state):
        opcode, value = state
        if opcode == re.sre_parse.LITERAL:
            return strats.just(hunichr(value))
        elif opcode == re.sre_parse.NOT_LITERAL:
            return strats.characters(blacklist_characters=hunichr(value))
        elif opcode == re.sre_parse.AT:
            return strats.just('')
        elif opcode == re.sre_parse.IN:
            if value[0][0] == re.sre_parse.NEGATE:
                candidates = []
                for v in value[1:]:
                    candidates.extend(chain(*(self._handle_character_sets(v))))
                return strats.characters(blacklist_characters=candidates)
            else:
                candidates = []
                for v in value:
                    candidates.extend(chain(*(self._handle_character_sets(v))))
                return strats.sampled_from(candidates)
        elif opcode == re.sre_parse.ANY:
            return strats.characters()
        elif opcode == re.sre_parse.BRANCH:
            branches = []
            for val in value[1]:
                branch = [self._handle_state(v) for v in val]
                branches.append(strategy_concat(branch))
            return strats.one_of(branches)
        elif opcode == re.sre_parse.SUBPATTERN:
            parts = []
            for part in value[1]:
                parts.append(self._handle_state(part))
            result = strategy_concat(parts)
            if value[0]:
                self.cache[value[0]] = result
                result = strats.shared(result, key=value[0])
            return result
        elif opcode == re.sre_parse.ASSERT:
            result = []
            for part in value[1]:
                result.append(self._handle_state(part))
            return strategy_concat(result)
        elif opcode == re.sre_parse.ASSERT_NOT:
            return strats.just('')
        elif opcode == re.sre_parse.GROUPREF:
            return strats.shared(self.cache[value], key=value)
        elif opcode == re.sre_parse.MIN_REPEAT:
            start_range, end_range, val = value
            result = []
            for v in val:
                part = strats.lists(
                    self._handle_state(v),
                    min_size=start_range,
                    max_size=end_range
                ).map(lambda x: u"".join(x))
                result.append(part)
            return strategy_concat(result)
        elif opcode == re.sre_parse.MAX_REPEAT:
            start_range, end_range, val = value
            result = []
            for v in val:
                part = strats.lists(
                    self._handle_state(v),
                    min_size=start_range,
                    max_size=end_range
                ).map(lambda x: u"".join(x))
                result.append(part)
            return strats.tuples(*result).map(lambda x: u"".join(x))
        else:
            import ipdb; ipdb.set_trace()
            raise NotImplementedError(opcode)


# From hypothesis.strategies.py
@defines_strategy
def regex(pattern):
#    from hypothesis.searchstrategy.regex import RegexStrategy
    return RegexStrategy(pattern)



from .tree import Node
from .support import is_key, is_leaf_key

def get_alias_nodes(node: Node):
    for c in node.children: # TODO this presume tree has Root
        if is_leaf_key(c):
            yield c

def remove_hidden_fields(node, language): # TODO should be like a plugin
    indicator = f'[{language} hide]'
    aliases = get_alias_nodes(node)
    to_remove = [x.value for x in aliases if indicator in x.annotation]
    return recursize_remove_hidden_fields(node, indicator, to_remove)


def apply_type_kind(node: Node, language='graphql'): # TODO should be like a plugin

    if f'[{language} input]' in node.annotation:
        node.is_input = True
    if f'[{language} interface]' in node.annotation:
        node.is_interface = True
    if not node.children:
        return
    for c in node.children:
        apply_type_kind(c, language)
    return node


def recursize_remove_hidden_fields(node: Node, indicator, aliases_to_remove):
    if not node.children:
        return
    for c in node.children:
        has_alias_to_remove = lambda x: (is_key(x) and x.children[0].value in aliases_to_remove)
        node.children = [x for x in node.children if not indicator in x.annotation and not has_alias_to_remove(x)]
        recursize_remove_hidden_fields(c, indicator, aliases_to_remove)
    return node



from genson import SchemaBuilder
import json
from ..from_jsonschema import from_jsonschema, SchemaBlock
from ..support import pretty

def replace_special_types(array):
    result = []
    inferrable_types = (dict, str, bool, int, float, list,)
    for obj in array:
        if isinstance(obj, (dict, list)):
            data = json.dumps(obj, default=str)
            result.append(json.loads(data))
        elif any([isinstance(obj, x) for x in inferrable_types]):
            result.append(obj)
    return result

def infer_schema(data_array, ):
    builder = SchemaBuilder()
    for data in data_array:
        builder.add_schema({"type": "object", "properties": {},})
        builder.add_object(data)
    schema = builder.to_schema()
    return schema

def infer_skema(data_array, ref_name='Root'):
    data_array = replace_special_types(data_array)
    schema = infer_schema(data_array)
    return from_jsonschema(schema, ref_name=ref_name)



from random import randint
from funcy import omit
from hypothesis import strategies as hs
# from hypothesis._strategies import from_regex as regex
import json
from .regex import regex


def gen_int(prop):
    min_value = prop.get("minimum", 0)
    max_value = prop.get("maximum", ) or prop.get("exclusiveMaximum", ) or (20 if not min_value else None)
    return hs.integers(min_value=min_value, max_value=max_value)


def gen_string(prop):
    min_value = prop.get("minLength", 5)
    max_value = prop.get("maxLength", 20)
    # pattern = prop.get('pattern', r'[ A-Za-z\d]+')
    # return regex(pattern, )\
    #         .filter(lambda x: min_value <= len(x) and len(x) <= max_value)\
    #         .map(lambda x: x.rstrip('\x00'))
    return hs.text(
        hs.characters(min_codepoint=97,max_codepoint=122, ),
        min_size=min_value, 
        max_size=max_value

            # blacklist_categories=('Cc', 'Cs')),
        ).map(lambda s: s.strip()
    ).filter(lambda s: len(s) > min_value)



def should_include(key, required_list):
    if key in required_list:
        return True
    else:
        return bool(randint(0, 1))


def gen_array(prop, customs):
    min_items = prop.get("minItems", 0)
    max_items = prop.get("maxItems", 5)
    if prop.get("items", {}):
        generator = get_generator(prop.get("items"), customs)
        return hs.lists(elements=generator,
                        min_size=min_items,
                        max_size=max_items)
    else:
        return hs.lists(elements=gen_anything(),
                        min_size=min_items,
                        max_size=max_items)


def gen_anything():
    return hs.one_of(gen_int({}), gen_string({}), hs.booleans(), hs.none())

def gen_json_values():
    return gen_string({}) | hs.booleans() | gen_int({}) | hs.none() | hs.floats(allow_nan=False, allow_infinity=False)

def gen_any_obj():
    return hs.recursive(hs.dictionaries(gen_string({}), gen_json_values()),
                        lambda children: hs.dictionaries(gen_string({}), children),
                        max_leaves=10)

def gen_object(prop, customs):

    required = prop.get("required", [])
    output = {}
    prop_key = "properties"
    if prop.get("properties", None) is None:
        if prop["additionalProperties"]:
            return gen_any_obj()

    for k in prop[prop_key].keys():
        json_prop = prop[prop_key][k]

        if should_include(k, required):
            output[k] = get_generator(json_prop, customs)


    return hs.fixed_dictionaries(output)


def gen_enum(prop):
    enum = prop["enum"]
    return hs.sampled_from(enum)

def gen_const(prop):
    const = prop["const"]
    return hs.just(const)

def gen_bool(prop):
    return hs.booleans()

def gen_one_of(prop, customs):
    possible_values = []
    for value in prop["oneOf"]:
        possible_values.append(get_generator(value, customs))

    return hs.one_of(possible_values)

def gen_any_of(prop, customs):
    possible_values = []
    for value in prop["anyOf"]:
        possible_values.append(get_generator(value, customs))

    return hs.one_of(possible_values)

specials = ['allOf', 'oneOf', 'anyOf']

def merge_all_of(prop,):
    dicts = prop["allOf"]

    if not all([any([x in d for x in specials]) or d.get('type') == 'object'  for d in dicts]):
        raise Exception(f'allOf is supported as array of object types: {json.dumps(dicts, indent=4, default=repr)}')
    result = {
        'additionalProperties': True,
        'required': [],
        'properties': {},
    }
    for obj in dicts:
        if 'allOf' in obj:
            obj = merge_all_of(obj)
    
        result['additionalProperties'] = result['additionalProperties'] and bool(obj.get('additionalProperties', True))
        result['properties'] = {**result['properties'], **(obj.get('properties', {}) or {})}
        result['required'] = result['required'] + (obj.get('required', [])  or [])
    result['required'] = list(set(result['required']))
    return result

def gen_all_of(prop, customs):
    result = merge_all_of(prop)
    return gen_object(result, customs)

def get_generator(prop, customs={}):
    disp = {"string": gen_string,
            "integer": gen_int,
            "number": gen_int,
            "boolean": gen_bool,
            
    }
    if prop.get('title') != None:
        title = prop.get('title', '') # .strip()
        if title in customs:
            return hs.builds(lambda x: customs[title](), hs.integers())
    if not omit(prop, ['title', 'description']):
        return gen_anything()

    enum = prop.get("enum", None)
    if enum is not None:
        return gen_enum(prop)

    if 'const' in prop:
        return gen_const(prop)

    one_of = prop.get("oneOf", None)
    if one_of is not None:
        return gen_one_of(prop, customs)

    any_of = prop.get("anyOf", None)
    if any_of is not None:
        return gen_any_of(prop, customs)

    all_of = prop.get("allOf", None)
    if all_of is not None:
        return gen_all_of(prop, customs)

    if 'object' == prop.get('type'):
        return gen_object(prop, customs)

    if 'array' == prop.get('type'):
        return gen_array(prop, customs)

    json_type = prop.get("type", None)
    if json_type is None:
        raise JsonTypeError("Couldnt find type in prop {0}".format(prop))

    return disp[json_type](prop)



class JsonTypeError(Exception):
    pass




from .. import to_jsonschema
from jsonschema import validate
import json
from typing import List
from hypothesis import given, settings
from .hypo_schema import get_generator






def fake_data(schema: str, ref=None, amount=5, resolvers={}, from_json=False) -> List[dict]:
    if from_json:
        json_schema=schema
    else:
        json_schema=to_jsonschema(schema, ref=ref, resolve=True)
    examples = []
    @given(get_generator(json_schema, resolvers))
    @settings(max_examples=amount)
    def generate(example_data):
        examples.append(example_data)
    generate()
    return examples





from collections import Counter
from skema import tokenize, make_tree
from funcy import rcompose, partial, tap
import operator
from skema.tree import copy, Node
from typing import List, Callable, Tuple
from skema.support import is_leaf_key, is_scalar
from skema.split_references import (FORBIDDEN_TYPE_NAMES,
                                    breadth_first_traversal, get_leaves,
                                    is_valid_as_reference,
                                    search_cascaded_name,
                                    remove_ellipses,
                                    remove_nulls,
                                    # dereference_objects_inside_lists,
                                    split_references,
                                    merge_ands,
                                    merge_scalar_unions,
                                    # replace_aliases,
                                    replace_types,
                                    get_alias_nodes,
                                    INTERFACE_END_KEYWORD,
                                    mark_enums,
                                    )
from skema.preprocess import remove_hidden_fields, apply_type_kind
from .support import (capitalize, is_and_key, is_enum_key, is_key, is_list_key,
                      is_object, is_or_key, is_scalar, is_leaf, is_leaf_key, is_and_object)


def remove_ands(refs): 
    refs_and_indexes = [merge_ands(r, refs) for r in refs]
    refs = [ref for ref, indexes in refs_and_indexes]
    # use them as implements
    indexes = set([i for ref, indexes in refs_and_indexes for i in indexes])
    indexes = sorted(indexes, reverse=True)
    interfaces = [copy(refs[i]) for i in indexes]
    for node in interfaces:
        node.value = node.value + INTERFACE_END_KEYWORD
        node.is_interface = True
    # i don't want to delete interfaces types
    # leaves = [leaf for r in refs for leaf in get_leaves(r) if r.value.lower() != 'root'] # TODO presume no root
    # for i in indexes:
    #     if not [l for l in leaves if l.value == refs[i].value]:
    #         refs.pop(i)
    return refs + interfaces

def search_non_existing_subgroup(subgroup: Tuple[str, ...], groups: List[Tuple[str, ...]]) -> Tuple[str, ...]:
    subgroup = ('', ) + subgroup
    while len(subgroup) > 1:
        if subgroup[1:] not in groups:
            subgroup = subgroup[1:]
        else:
            # print('FOUND', subgroup)
            break
    return subgroup

def search_existing_subgroup(subgroup: Tuple[str, ...], groups: List[Tuple[str, ...]]) -> Tuple[str, ...]:
    subgroup = ('', ) + subgroup
    while len(subgroup) > 1:
        if subgroup not in groups:
            subgroup = subgroup[1:]
        else:
            # print('FOUND', subgroup)
            break
    # print('subgroup', subgroup)
    return subgroup

def remove_parent_names(refs: List[Node]) -> List[Node]:
    refs = sorted(refs, key=lambda v: len(v.value))
    ref_names = [n.value if isinstance(n.value, tuple) else (n.value,) for n in refs]
    if not len(set(ref_names)) == len(ref_names):
        duplicates = [x for x, count in Counter(ref_names).items() if count > 1]
        print(f'WARNING, found a duplicate: {duplicates}')
    for ref in refs:
        ref.value = ref.value if isinstance(ref.value, tuple) else (ref.value,)
        other_ref_names = [r.value for r in refs if r.value != ref.value]
        subgroup = search_non_existing_subgroup(ref.value, other_ref_names)
        ref.value = subgroup
        # print('ref.value', ref.value)
    # print([r.value for r in refs])
    return [compute_new_names(ref, [r.value for r in refs]) for ref in refs]

def compute_new_names(node: Node, ref_names: List[str]) -> Node:
    leaves = get_leaves(node)
    leaves_to_compute = [l for l in leaves if isinstance(l.value, tuple)]
    # print('leaves_to_compute', leaves_to_compute)
    for leaf in leaves_to_compute:
        subgroup = search_existing_subgroup(leaf.value, ref_names)
        leaf.value = subgroup
    return node
        
def remove_tuples(refs: List[Node]) -> List[Node]:
    for ref in refs:
        nodes = breadth_first_traversal(ref)
        for node in nodes:
            if isinstance(node.value, tuple):
                node.value = ''.join(node.value)
    return refs







def split_tree_parts(string, language, hide=[], only=None, is_valid_as_reference=is_valid_as_reference, ) -> List[Node]:
    preprocess_refs = rcompose(
        remove_parent_names,
        # mark_enums,
        remove_tuples,
        remove_ands,
        lambda refs: merge_scalar_unions(refs,),
        lambda refs: [replace_types(r, refs) for r in refs],
        partial(filter, lambda x: x.value.lower() != 'root'),
        list,
    )
    node = make_tree(tokenize(string))
    node = remove_ellipses(node)
    node = remove_nulls(node)
    node = remove_hidden_fields(node, language)
    node = apply_type_kind(node, language)
    # node = replace_aliases(node)
    # print(node)
    refs = [*get_alias_nodes(node)] + [*split_references(node, is_valid_as_reference)]
    refs = preprocess_refs(refs)
    if only and isinstance(only, list):
        refs = [c for c in refs if c.value in only]
    if hide and isinstance(hide, list):
        refs = [c for c in refs if c.value not in hide]
    return refs
    # refs = [r for r in refs if not (is_leaf_key(r) and r.value in scalar_already_present)]




import os.path
import sys
import json
import skema
import subprocess
import os, random
from ..to_graphql import to_graphql
from ..to_python import to_python

class temporary_write:
        def __init__(self,  data, path=str(random.random())[3:]):
                self.path = os.path.abspath(path)
                self.data = data
                f = open(self.path, 'w')
                f.write(self.data)
                f.close()

        __enter__ = lambda self: self.aquire()
        aquire = lambda self: self.path
        __exit__ = lambda self, a, b, c: self.release()
        release = lambda self: os.remove(self.path)


def get_result_file(skema_path, extension):
    skema_path = os.path.abspath(skema_path)
    result_dir = os.path.dirname(skema_path)
    name, _ = os.path.splitext(skema_path)
    name = name.split('/')[-1]
    result_file = os.path.join(result_dir, name + extension)
    # print(result_file)
    return result_file

def generate_types(skema_path, extension, args, result_file=None, ref=None, hide=[], only=None):
    with open(skema_path) as f:
        data = f.read()
        json_schema = skema.to_jsonschema(data, hide=hide, only=only)
    root_name = json_schema.get('$ref').split('/')[-1]
    json_schema = json.dumps(json_schema, indent=4, default=str)
    # print(json_schema)
    result_file = result_file or get_result_file(skema_path, extension)
    with temporary_write(json_schema) as schema_path:
        command = f'quicktype "{schema_path}#/definitions/{ref or ""}" -o {result_file} --top-level {ref or root_name} --src-lang schema {args}'
        print(command)
        try:
            res = subprocess.run(command, shell=True, check=True, stdout=subprocess.PIPE)
            print(res.stdout.decode())
            return result_file
        except subprocess.CalledProcessError as e:
            print(e.stderr)
            raise Exception('quicktype failed')

def generate_graphql(skema_path, result_file=None, hide=[], only=None):
    with open(skema_path) as f:
        data = f.read()
    schema = to_graphql(data, hide=hide, only=only)
    result_file = result_file or get_result_file(skema_path, '.graphql')
    with open(result_file, 'w') as f:
        f.write(schema)

def generate_python(skema_path, result_file=None, hide=[], only=None):
    with open(skema_path) as f:
        data = f.read()
    schema = to_python(data, hide=hide, only=only)
    result_file = result_file or get_result_file(skema_path, '.py')
    with open(result_file, 'w') as f:
        f.write(schema)

def generate_jsonschema(skema_path, result_file=None, ref=None, resolve=False, hide=[], only=None):
        result_path = result_file or get_result_file(skema_path, '.json')
        with open(skema_path) as f:
                data = f.read()
        schema = skema.to_jsonschema(data, ref=ref, resolve=resolve, hide=hide, only=only)
        with open(result_path, 'w') as f:
                f.write(json.dumps(schema, indent=4))
    



import json



def _resolve(schema, definitions, add_titles=False):
    if not isinstance(schema, dict):
        return
    for k, v in schema.items():
        # print(k)
        if isinstance(v, dict) and '$ref' in list(v.keys()):
            # print('_')
            ref = v['$ref'].split('/')[-1]
            schema[k] = definitions[ref]
            if add_titles:
                schema[k].update({'title': ref})
        elif isinstance(v, dict):
            _resolve(schema[k], definitions, add_titles)
        elif isinstance(v, list):
            for i, item in enumerate(v):
                if isinstance(item, dict) and '$ref' in list(item.keys()):
                    # print('_')
                    ref = item['$ref'].split('/')[-1]
                    schema[k][i] = definitions[ref]
                elif isinstance(item, dict):
                    _resolve(schema[k][i], definitions, add_titles)
                else:
                    value = json.dumps(v, indent=4)[:400] + "\n"
                    # print(f'should not be here, {k}={value}')
        else:
            value = json.dumps(v, indent=4)[:400] + "\n"
            # print(f'should not be here, {k}={value}')


def resolve_refs(schema, add_titles=False):
    if not 'definitions' in schema:
        return
    definitions = schema['definitions']
    _resolve(schema, definitions, add_titles)
    if '$ref' in schema:
        ref = schema['$ref'].split('/')[-1]
        schema.update(schema['definitions'][ref])
        if add_titles:
            schema.update({'title': ref})
        del schema['$ref']
    del schema['definitions']
    return


if __name__ == '__main__':
    schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$ref": "#/definitions/C",
        "definitions": {
            "X": {
                "title": "X",
                "type": "object",
                "required": [
                    "ciao",
                    "b",
                    "c"
                ],
                "properties": {
                    "ciao": {
                        "type": "string"
                    },
                    "b": {
                        "oneOf": [
                            {
                                "type": "number",
                                "multipleOf": 1.0
                            },
                            {
                                "type": "string"
                            }
                        ]
                    },
                    "c": {
                        "const": "ciao"
                    }
                },
                "additionalProperties": False
            },
            "B": {
                "title": "B",
                "type": "object",
                "required": [
                    "x"
                ],
                "properties": {
                    "x": {
                        "type": "number",
                        "multipleOf": 1.0
                    }
                },
                "additionalProperties": False
            },
            "C": {
                "title": "C",
                "type": "object",
                "required": [
                    "x",
                    "b"
                ],
                "properties": {
                    "x": {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/X"
                            },
                            {
                                "$ref": "#/definitions/B"
                            }
                        ]
                    },
                    "b": {
                        "$ref": "#/definitions/B"
                    }
                },
                "additionalProperties": False
            }
        }
    }

    resolve_refs(schema,)
    # print(json.dumps(schema, indent=4))